#if !defined(NDEBUG) && defined(MEM_CHECK)

#define MEM_H /* We don't want mem.h included, it's defines will screw up everything here. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "common.h"
#include "assert.h"
#include "log.h"

static const unsigned char padding [128] =
                           "-padding-|-padding-|-padding-|"
                           "-padding-|-padding-|-padding-|"
                           "-padding-|-padding-|-padding-|"
                           "-padding-|-padding-|-padding-|"
                           "-padding";
                           /* This sequence of bytes is put at the start and end of allocated
                              blocks of data. It's a good idea to make it a multiple of 64 so
                              that it doesn't mess up alignment. */

unsigned int cur_memory = 0;
unsigned int peak_memory = 0;

struct meminfo
 {
  unsigned char *pos;
  unsigned int size;
  const char *file, *function;
  unsigned int line;
 } *start = 0, *end = 0, *alloc_end = 0;

int registerd_shutdown = 0;

static void memShutdown(void)
 {
  if(start != end)
   {
    unsigned int total = 0;
    
    logMessage(Warning, "%lu memory leak%s detected, as follows:", end - start, end - start == 1 ? "":"s");
    
    for(; start != end; --end)
     {
      total += end[-1].size;
      logMessage(Warning, "%s %s from %p.", sourceString(end[-1].file, end[-1].function, end[-1].line), memoryString(end[-1].size), end[-1].pos);
      
      if(memcmp(end[-1].pos - sizeof padding, padding, sizeof padding))
       logMessage(Warning, "%s Damage detected before memory allocated at %p.", sourceString(end[-1].file, end[-1].function, end[-1].line), end[-1].pos);
      
      if(memcmp(end[-1].pos + end[-1].size, padding, sizeof padding))
       logMessage(Warning, "%s Damage detected after memory allocated at %p.", sourceString(end[-1].file, end[-1].function, end[-1].line), end[-1].pos);
      
      free(end[-1].pos);
     }
    
    logMessage(Warning, "Total %s of leaked memory.", memoryString(total));
    
    if(cur_memory != total)
     logMessage(Error, "Marked more memory as used than is actually used; internal error?");
   }
  else if(cur_memory)
   logMessage(Error, "Marked more memory as used than is actually used; internal error?");
  
  logMessage(Notice, "Most memory allocated one time: %s", memoryString(peak_memory));
  
  peak_memory = 0;
  cur_memory = 0;
  
  if(start)
   {
    free(start);
    start = end = alloc_end = 0;
   }
  
  registerd_shutdown = 0;
 }

int checkDamage(struct meminfo *info, const char *file, const char *function, unsigned int line)
 {
  int okay = 1;
  
  if(info < start || info >= end)
   {
    logMessage(Warning, "%s checkDamage passed invalid info; internal error?", sourceString(file, function, line));
    return 0;
   }
  
  if(memcmp(info->pos - sizeof padding, padding, sizeof padding))
   {
    logMessage(Warning, "%s Damage detected before memory allocated at %p.", sourceString(file, function, line), info->pos);
    logMessage(Warning, "%s Memory from previous error was allocated here.", sourceString(info->file, info->function, info->line));
    memcpy(info->pos - sizeof padding, padding, sizeof padding);
    okay = 0;
   }
   
  if(memcmp(info->pos + info->size, padding, sizeof padding))
   {
    logMessage(Warning, "%s Damage detected after memory allocated at %p.", sourceString(file, function, line), info->pos);
    logMessage(Warning, "%s Memory from previous error was allocated here.", sourceString(info->file, info->function, info->line));
    memcpy(info->pos + info->size, padding, sizeof padding);
    okay = 0;
   }
  
  return okay;
 }

static struct meminfo* recallRegion(unsigned char *data)
 {
  struct meminfo *s = start, *e = end, *p = s+(e-s)/2;
  
  for(; s < e; p = s+(e-s)/2)
   if(p->pos - sizeof padding > data)
    e = p;
   else if(p->pos + p->size + sizeof padding < data)
    s = p+1;
   else return p;
  
  return 0;
 }

static struct meminfo* rememberChunk(unsigned char *data)
 {
  struct meminfo *s, *e, *p;
  
  /* make sure we have enough room to add a new chunk. */
  if(end == alloc_end)
   {
    if(start == alloc_end)
     {
      if(start != 0) /* I doubt this would ever be the case, but just in case. . . */
       free(start);
      
      if(!(start = end = (meminfo *)malloc(sizeof *start * 16)))
       {
        logMessage(Notice, "Not enough memory to reallocate memory chunk list.");
        return 0;
       }
      
      alloc_end = start+16;
     }
    else
     {
      struct meminfo *old = start;
      if(!(start = realloc(start, 2*(alloc_end - start)*sizeof *start)))
       {
        start = old;
        logMessage(Notice, "Not enough memory to reallocate memory chunk list.");
        return 0;
       }
     
      end = end - old + start;
      alloc_end = 2*(alloc_end - old) + start;
     }
     
    logMessage(Notice, "Memory chunk list expanded; can now hold %d chunks.", alloc_end - start);
   }
   
  s = start, e = end, p = s+(e-s)/2;
  
  for(; s < e; p = s+(e-s)/2)
   if(p->pos > data)
    e = p;
   else if(p->pos < data)
    s = p+1;
   else
    {
     logMessage(Error, "Chunk already exists! Internal error?");
     return 0;
    }
  
  memmove(p+1, p, (end-p)*sizeof *p);
  p->pos = data;
  p->size = 0;
  p->file = p->function = 0;
  p->line = 0;
  ++end;
  return p;
 }
 
static int forgetChunk(struct meminfo *chunk)
 {
  if(chunk < start && chunk >= end)
   {
    logMessage(Error, "Invalid chunk! Internal error?");
    return 0;
   }
  
  --end;
  memmove(chunk, chunk+1, (end-chunk)*sizeof *chunk);
  
  return 1;
 }

#ifdef PARANOID_MEMORY
void verifyIntegrety(const char *file, const char *function, unsigned int line)
 {
  struct meminfo *i = start;
  unsigned int total = 0;
  
  for(; i != end; ++i)
   {
    if(i != start)
     {
      if(i->pos <= i[-1].pos+i[-1].size+sizeof padding)
       {
        logMessage(Error, "%s pointers %p and %p overlap or are out of order! Internal Error?", sourceString(file, function, line), i->pos, i[-1].pos);
        logMessage(Warning, "%s Memory from previous error was allocated here.", sourceString(i->file, i->function, i->line));
        logMessage(Warning, "%s Memory from previous error was allocated here.", sourceString(i[-1].file, i[-1].function, i[-1].line));
       }
     }
    
    checkDamage(i, file, function, line);
    total += i->size;
   }
  
  if(total != cur_memory)
   {
    logMessage(Error, "%s Recorded memory used not equal to calculated memory used! Internal Error?", sourceString(file, function, line));
    cur_memory = total;
   }
 }
#endif

int verifyTrace(void * data, const char *file, const char *function, unsigned int line)
 {
  struct meminfo *info;
  
  if(!data)
   {
    logMessage(Warning, "%s Pointer %p is null.", sourceString(file, function, line), data);
    return 0;
   }
   
  if(!(info = recallRegion(data)))
   {
    logMessage(Warning, "%s Pointer %p doesn't appear to have been allocated.", sourceString(file, function, line), data);
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return 0;
   }
   
  if(info->pos != data)
   {
    logMessage(Warning, "%s Pointer %p is not start of region %p-%p.", sourceString(file, function, line), data, info->pos, info->pos+info->size);
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return 0;
   }
  
  if(!checkDamage(info, file, function, line))
   {
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return 0;
   }
  
  #ifdef PARANOID_MEMORY
  verifyIntegrety(file, function, line);
  #endif
  return 1;
 }

void * mallocTrace(unsigned int size, const char *file, const char *function, unsigned int line)
 {
  unsigned char * data;
  struct meminfo *info;
  
  if(!registerd_shutdown)
   {
    atexit(&memShutdown);
    registerd_shutdown = 1;
   }
  
  if(size == 0)
   {
    logMessage(Notice, "%s Attempted to allocate 0 bytes; returning 0.", sourceString(file, function, line));
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return 0;
   }
   
  if(!(data = malloc(size + 2 * sizeof padding)))
   {
    logMessage(Notice, "%s Out of memory; returning 0.", sourceString(file, function, line));
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return 0;
   }
  
  if(!(info = rememberChunk(data + sizeof padding)))
   {
    logMessage(Notice, "%s Error remembering chunk; freeing data and returning 0.", sourceString(file, function, line));
    free(data);
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return 0;
   }
  
  info->size = size;
  info->file = file;
  info->function = function;
  info->line = line;
  
  if((cur_memory += size) > peak_memory)
   peak_memory = cur_memory;
  
  memcpy(data, padding, sizeof padding);
  memset(data+sizeof padding, 0xff, size);
  memcpy(data+size+sizeof padding, padding, sizeof padding);

  #ifdef PARANOID_MEMORY
  verifyIntegrety(file, function, line);
  #endif
  
  return info->pos;
 }

void freeTrace(void * data, const char *file, const char *function, unsigned int line)
 {
  struct meminfo *info;
  
  if(!data)
   {
    logMessage(Warning, "%s Attempted to free %p.", sourceString(file, function, line), data);
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return;
   }
   
  if(!(info = recallRegion(data)))
   {
    logMessage(Warning, "%s Attempted to free %p which doesn't appear to have been allocated. Maybe you freed it twice?", sourceString(file, function, line), data);
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return;
   }
   
  if(info->pos != data)
   {
    logMessage(Warning, "%s Attempted to free %p from allocated region %p-%p, this is not the correct pointer to use to free that region.", sourceString(file, function, line), data, info->pos, info->pos+info->size);
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return;
   }
  
  checkDamage(info, file, function, line);
  
  free(info->pos - sizeof padding);
  
  cur_memory -= info->size;
  
  if(!forgetChunk(info))
   {
    logMessage(Error, "Unable to forget chunk; internal error?");
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return;
   }
  
  #ifdef PARANOID_MEMORY
  verifyIntegrety(file, function, line);
  #endif
 }
 
void * reallocTrace(void *data, unsigned int size, const char *file, const char *function, unsigned int line)
 {
  struct meminfo *info;
  
  if(!data)
   {
    logMessage(Notice, "%s Attempted to resize a null pointer; passing parameters to malloc_trace instead.", sourceString(file, function, line));
    return mallocTrace(size, file, function, line);
   }
   
  if(!(info = recallRegion(data)))
   {
    logMessage(Warning, "%s Attempted to resize data at %p which doesn't appear to be allocated.", sourceString(file, function, line), data);
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return 0;
   }
   
  if(!size)
   {
    logMessage(Warning, "%s Attempted to resize data at %p to %s.", sourceString(file, function, line), data, memoryString(size));
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return 0;
   }
   
  if(info->pos != data)
   {
    logMessage(Warning, "%s Attempted to resize %p from allocated region %p-%p, this is not the correct pointer to use to resize that region.", sourceString(file, function, line), data, info->pos, info->pos+info->size);
    logMessage(Warning, "%s Memory from previous error was allocated here.", sourceString(info->file, info->function, info->line));
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return 0;
   }
  
  checkDamage(info, file, function, line);
   
  if(info->size == size)
   {
    logMessage(Notice, "%s Attempted to resize chunk at %p to original size, doing nothing.", sourceString(file, function, line), info->pos);
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return info->pos;
   }
   
  if(!(data = realloc(info->pos - sizeof padding, size + 2 * sizeof padding)))
   {
    logMessage(Notice, "%s Not enough memory to resize region; returning 0.", sourceString(file, function, line));
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return 0;
   }
  
  cur_memory -= info->size;
  if((cur_memory += size) > peak_memory)
   peak_memory = cur_memory;
   
  if(size > info->size)
   {
    /* clear the extra memory to some recognizable pattern. */
    memset((unsigned char*)data + sizeof padding + info->size, 0xff, size - info->size);
   }
   
  memcpy((unsigned char*)data + sizeof padding + size, padding, sizeof padding);
  
  if(!(forgetChunk(info)))
   {
    logMessage(Error, "Unable to forget chunk; internal error?");
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return data+sizeof padding;
   }
  
  if(!(info = rememberChunk((unsigned char*)data + sizeof padding)))
   {
    logMessage(Error, "Unable to remember chunk; internal error? "
                      "We just had the memory, it couldn't have gone anywhere!");
    #ifdef PARANOID_MEMORY
    verifyIntegrety(file, function, line);
    #endif
    return data+sizeof padding;
   }
  
  info->size = size;
  info->file = file;
  info->function = function;
  info->line = line;
  
  #ifdef PARANOID_MEMORY
  verifyIntegrety(file, function, line);
  #endif
  return info->pos;
 }

#else
/* Prevent the source file from being empty. */
typedef int unused;
#endif
